@@ -8,13 +8,14 @@ from skimage.transform import rescale
 
 
 class ims:
-    def __init__(self, file, ResolutionLevelLock=0, cache_location=None, mem_size=None, disk_size=2000):
+    def __init__(self, file, ResolutionLevelLock=0, write=False, cache_location=None, mem_size=None, disk_size=2000, squeeze_output=False):
         
         ##  mem_size = in gigabytes that remain FREE as cache fills
         ##  disk_size = in gigabytes that remain FREE as cache fills
         ## NOTE: Caching is currently not implemented.  
         
         self.filePathComplete = file
+        self.write = write
         self.open()
         self.filePathBase = os.path.split(file)[0]
         self.fileName = os.path.split(file)[1]
@@ -27,9 +28,11 @@ class ims:
         self.disk_size = disk_size * 1e9 if disk_size is not None else None
         self.mem_size = mem_size * 1e9 if mem_size is not None else None
         self.memCache = {}
+        self.squeeze_output = squeeze_output
         self.cacheFiles = []
         self.metaData = {}
         self.ResolutionLevelLock = ResolutionLevelLock
+        
 
         resolution_0 = self.dataset['ResolutionLevel 0']
         time_point_0 = resolution_0['TimePoint 0']
@@ -127,16 +130,25 @@ class ims:
     #     self.hf = None
         
     def open(self):
-        print('Opening file: {} \n'.format(self.filePathComplete))
-        self.hf = h5py.File(self.filePathComplete, 'r', swmr=True)
-        self.dataset = self.hf['DataSet']
-        # print('OPENED file: {} \n'.format(self.filePathComplete))
+        if self.write == False:
+            print('Opening readonly file: {} \n'.format(self.filePathComplete))
+            self.hf = h5py.File(self.filePathComplete, 'r', swmr=True)
+            self.dataset = self.hf['DataSet']
+            # print('OPENED file: {} \n'.format(self.filePathComplete))
+        elif self.write == True:
+            print('Opening writeable file: {} \n'.format(self.filePathComplete))
+            self.hf = h5py.File(self.filePathComplete, 'a', swmr=True)
+            self.dataset = self.hf['DataSet']
+            # print('OPENED file: {} \n'.format(self.filePathComplete))
     
     def __del__(self):
         self.close()
     
     def close(self):
         ## Implement flush?
+        if self.write == True:
+            print('Flushing Buffers to Disk')
+            self.hf.flush()
         print('Closing file: {} \n'.format(self.filePathComplete))
         if self.hf is not None:
             self.hf.close()
@@ -157,10 +169,50 @@ class ims:
         will be obeyed.  If ResolutionLevelLock is == None - default resolution is 0 (full-res)
         and a slice of 5 or less dimensions will extract information from resolutionLevel 0.
 
-        ResolutionLevelLock is used when building a multi-resolution series to load into napari
-        This option enables a 5D slice to lock on to a specified resolution level.
+        ResolutionLevelLock is used to focus on a specific resoltion level.
+        This option enables a 5D slicing to lock on to a specified resolution level.
         """
 
+        res, key = self.transform_key(key)        
+
+        slice_returned = self.get_slice(
+            r=res if res is not None else 0,  # Force ResolutionLock of None to be 0 when slicing
+            t=self.slice_fixer(key[0], 't', res=res),
+            c=self.slice_fixer(key[1], 'c', res=res),
+            z=self.slice_fixer(key[2], 'z', res=res),
+            y=self.slice_fixer(key[3], 'y', res=res),
+            x=self.slice_fixer(key[4], 'x', res=res)
+        )
+        return slice_returned
+
+    
+    def __setitem__(self,key,newValue):
+        
+        if self.write == False:
+            print("""
+                  IMS File can not be written to.
+                  imsClass.write = False.
+                  """)
+            pass
+        
+        res, key = self.transform_key(key)
+        
+        self.set_slice(
+            r=res if res is not None else 0,  # Force ResolutionLock of None to be 0 when slicing
+            t=self.slice_fixer(key[0], 't', res=res),
+            c=self.slice_fixer(key[1], 'c', res=res),
+            z=self.slice_fixer(key[2], 'z', res=res),
+            y=self.slice_fixer(key[3], 'y', res=res),
+            x=self.slice_fixer(key[4], 'x', res=res),
+            newData=newValue
+        )
+        
+        
+        
+        
+
+    def transform_key(self,key):
+        
         res = self.ResolutionLevelLock
 
         if not isinstance(key, slice) and not isinstance(key, int) and len(key) == 6:
@@ -183,16 +235,9 @@ class ims:
         while len(key) < 5:
             key.append(slice(None))
         key = tuple(key)
-
-        slice_returned = self.get_slice(
-            r=res if res is not None else 0,  # Force ResolutionLock of None to be 0 when slicing
-            t=self.slice_fixer(key[0], 't', res=res),
-            c=self.slice_fixer(key[1], 'c', res=res),
-            z=self.slice_fixer(key[2], 'z', res=res),
-            y=self.slice_fixer(key[3], 'y', res=res),
-            x=self.slice_fixer(key[4], 'x', res=res)
-        )
-        return slice_returned
+        
+        return res,key
+        
 
     def read_numerical_dataset_attr(self, attrib):
         return float(self.read_attribute('DataSetInfo/Image', attrib))
@@ -319,14 +364,48 @@ class ims:
         The napari_imaris_loader currently hard codes os.environ["NAPARI_ASYNC"] == '1'
         """
 
-        if "NAPARI_ASYNC" in os.environ and os.environ["NAPARI_ASYNC"] == '1':
-            output_array = np.squeeze(output_array)
-            print('Slices Requested: {} / Shape returned: {} \n'.format(incomingSlices,output_array.shape))
-            return output_array
-        # elif "NAPARI_OCTREE" in os.environ and os.environ["NAPARI_OCTREE"] == '1':
+        # if "NAPARI_ASYNC" in os.environ and os.environ["NAPARI_ASYNC"] == '1':
+        #     # output_array = np.squeeze(output_array)
+        #     print('Slices Requested: {} / Shape returned: {} \n'.format(incomingSlices,output_array.shape))
         #     return output_array
-        else:
+        # # elif "NAPARI_OCTREE" in os.environ and os.environ["NAPARI_OCTREE"] == '1':
+        # #     return output_array
+        # else:
+        if self.squeeze_output:
             return np.squeeze(output_array)
+        else:
+            return output_array
+
+    
+    def set_slice(self, r, t, c, z, y, x, newData):
+        """
+        IMS stores 3D datasets ONLY with Resolution, Time, and Color as 'directory'
+        structure writing HDF5.  Thus, data access can only happen across dims XYZ
+        for a specific RTC.
+        """
+
+        # incomingSlices = (r,t,c,z,y,x)
+        t_size = list(range(self.TimePoints)[t])
+        c_size = list(range(self.Channels)[c])
+        z_size = len(range(self.metaData[(r, 0, 0, 'shape')][-3])[z])
+        y_size = len(range(self.metaData[(r, 0, 0, 'shape')][-2])[y])
+        x_size = len(range(self.metaData[(r, 0, 0, 'shape')][-1])[x])
+
+        # if isinstance(newData,int):
+        toWrite = np.zeros((len(t_size), len(c_size), z_size, y_size, x_size), dtype=self.dtype)
+        toWrite[:] = newData
+        print(toWrite.shape)
+        print(toWrite)
+
+        print(t_size)    
+        print(t_size)
+        for idxt, t in enumerate(t_size):
+            for idxc, c in enumerate(c_size):
+                ## Below method is faster than all others tried
+                d_set_string = self.location_generator(r, t, c, data='data')
+                self.hf[d_set_string].write_direct(toWrite, np.s_[idxt, idxc, :, :, :], np.s_[z, y, x])
+
+
 
     def dtypeImgConvert(self, image):
         """
@@ -434,6 +513,7 @@ class ims:
         rescaleFactor = tuple([round(x/y,5) for x,y in zip(workingVolumeResolution,output_resolution)])
         print('Rescale Factor = {}'.format(rescaleFactor))
 
+        workingVolume = img_as_float32(workingVolume)
         workingVolume = rescale(workingVolume, rescaleFactor, anti_aliasing=anti_aliasing)
 
         return self.dtypeImgConvert(workingVolume)
@@ -490,6 +570,8 @@ class ims:
                 0,self.metaData[(resolutionLevel, 0, 0, 'shape')][-1]
             )
 
+        
+        failed = []
         for time in time_points:
             for color in channels:
                 for layer in range(self.metaData[(resolutionLevel,0,0,'shape')][-3]):
@@ -497,6 +579,54 @@ class ims:
                     if os.path.exists(fileName):
                         print('Skipping {} becasue it already exists'.format(fileName))
                         continue
-                    array = self[resolutionLevel,time,color,layer,cropYX[0]:cropYX[1],cropYX[2]:cropYX[3]]
+                    try:
+                        array = self[resolutionLevel,time,color,layer,cropYX[0]:cropYX[1],cropYX[2]:cropYX[3]]
+                    except:
+                        failed.append((resolutionLevel,time,color,layer,cropYX[0],cropYX[1],cropYX[2],cropYX[3]))
+                        continue
                     print('Saving: {}'.format(fileName))
-                    io.imsave(fileName, array)
+                    io.imsave(fileName, array, check_contrast=False)
+        
+        if len(failed) > 0:
+            print('Failed to extract the following layers:')
+            print(failed)
+        else:
+            print('All layers have been extracted')
+
+    def save_multilayer_tiff_stack(self, location=None, time_point=0, channel=0, resolution_level=0):
+        """
+        Extract 1 time point, 1 channel, all (z,y,x) for a specified resolution level.
+
+        location - path to the output tiff file. If not provided, file is saved next to the .ims file
+        Output: tiff stack (all z layers in one file)
+        """
+        if not location:
+            location = os.path.join(self.filePathBase, '{}_multilayer_stack.tif'.format(self.fileName))
+        elif type(location) != str or location == '':
+            raise TypeError('location must be a nonempty string')
+
+        if os.path.exists(location):
+            raise OSError('file at specified location already exists')
+        else:
+            os.makedirs(os.path.basename(location), exist_ok=True)
+
+        try:
+            time_point = int(time_point)
+            channel = int(channel)
+            resolution_level = int(resolution_level)
+        except TypeError:
+            raise TypeError('time_point, channel and resolution_level must be convertable to int')
+
+        try:
+            output_array = self[resolution_level, time_point, channel, :, :, :]
+        except:  # try extracting z layers one-by-one
+            output_array_shape = self.metaData[(resolution_level, time_point, channel, 'shape')][-3:]
+            output_array = np.zeros([*output_array_shape], dtype=self.dtype)  # layers with errors will be left as 0
+            for layer in range(output_array_shape[0]):
+                try:
+                    z_plane = self[resolution_level, time_point, channel, layer, :, :]
+                    output_array[layer, :, :] = z_plane
+                except Exception as e:
+                    print(f'Failed to extract layer {layer}: {e}')
+
+        io.imsave(location, output_array, check_contrast=False)
